% Example for robust CSP: the linear mixing model

clear all; close all
DO_FEATURE_EX = 0; % feature extraction test if true

addpath ./rcsp/
rng(0); 

% data gen
[V1, lam1, Sbar1, V2, lam2, Sbar2, Tst1, Tst2, Train1, Train2] = datagen();
delta = 6; % perturbation level, large value may lead to nonsmooth case 

% ++++++++++++++++++++++++++++++
% --- Run situation 1 
% ++++++++++++++++++++++++++++++
[V,E] = eig(Sbar2,Sbar1); 
[~,idx] = min(diag(E));
z0 = V(:,idx); z0 = z0/norm(z0);
GMAT2d = @(w) GMAT2(w, lam2, V2, delta, Sbar2);
HMAT2d = @(w) HMAT2(w, lam1, V1, delta, Sbar1);
[z1, RR1, ~, ~, resd1] = scfls(GMAT2d, HMAT2d, z0);
[z1t, RR1t, ~, ~, resd1t] = scfsls(GMAT2d, HMAT2d, z0);

% check eig order
[AA,TG] = GMAT2d(z1);
[BB,TH] = HMAT2d(z1);
disp(['1st   2ed', '   rho=',num2str(RR1(end))]);
disp([sort(eig(AA, BB)), sort(eig(AA+TG, BB+TH))]);

% ++++++++++++++++++++++++++++++
% --- Run situation 2 
% ++++++++++++++++++++++++++++++
[~,idx] = max(diag(E));
z0 = V(:,idx); z0 = z0/norm(z0);
GMAT2d = @(w) GMAT2(w, lam1, V1, delta, Sbar1);
HMAT2d = @(w) HMAT2(w, lam2, V2, delta, Sbar2);
[z2, RR2, ~, ~, resd2] = scfls(GMAT2d, HMAT2d, z0);
[z2t, RR2t, ~, ~, resd2t] = scfsls(GMAT2d, HMAT2d, z0);

% check eig order
[AA,TG] = GMAT2d(z2);
[BB,TH] = HMAT2d(z2);
disp(['1st   2ed', '   rho=',num2str(RR2(end))]);
disp([sort(eig(AA, BB)), sort(eig(AA+TG, BB+TH))]);

% ++++++++++++++++++++++++++++++
% --- Run simple iter
% ++++++++++++++++++++++++++++++
GMAT = @(w) Sbar2 + optmat_build(w, lam2, V2, delta);
HMAT = @(w) NegTrunc(Sbar1 - optmat_build(w, lam1, V1, delta));
[~,idx] = min(diag(E));
z0 = V(:,idx); z0 = z0/norm(z0);
[z1s, RR1s, ~] = scfitr(GMAT, HMAT, z0);

GMAT = @(w) Sbar1 + optmat_build(w, lam1, V1, delta);
HMAT = @(w) NegTrunc(Sbar2 - optmat_build(w, lam2, V2, delta));
[~,idx] = max(diag(E));
z0 = V(:,idx); z0 = z0/norm(z0);
[z2s, RR2s, ~] = scfitr(GMAT, HMAT, z0);

% ++++++++++++++++++++++++++++++
% --- Plot Rayleigh ratio, and error
% ++++++++++++++++++++++++++++++
%figure;
%subplot(1,2,1); plot(RR1, '-+'); 
%subplot(1,2,2); plot(RR2, '-or');
%xlabel('its'); ylabel('ratio');
%title('SCF NEP2');
%
%figure;
%subplot(1,2,1); plot(RR1s, '-+'); 
%subplot(1,2,2); plot(RR2s, '-or');
%xlabel('its'); ylabel('ratio');
%title('Simple itrs');

figure
plot(RR1s,'--+','LineWidth',2,'MarkerSize',8);
hold on;
plot(RR1t,'--xk','LineWidth',2,'MarkerSize',8);
plot(RR1,'--or','LineWidth',2,'MarkerSize',8);
xlabel('iterations','Fontsize',10);
ylabel('RQ values \rho(z_k)','Fontsize',10);
h_legend=legend('simple iter', 'Alg.1', 'Alg.2');
set(h_legend,'FontSize',10);
xlim([0,15]);
ylim([0.8,4]);
title('Situation 1');

figure
semilogy(abs(RR1t-RR1t(end)),'--xk','LineWidth',2,'MarkerSize',8);
hold on;
semilogy(abs(RR1-RR1(end)),'--or','LineWidth',2,'MarkerSize',8);
xlabel('iterations','Fontsize',10);
ylabel('|\rho(z_k) -\rho(z_*)|','Fontsize',10);
h_legend=legend('Alg.1', 'Alg.2');
xlim([0,15]);
set(h_legend,'FontSize',10);
title('Situation 1');


figure
plot(RR2s,'--+','LineWidth',2,'MarkerSize',8);
hold on;
plot(RR2t,'--xk','LineWidth',2,'MarkerSize',8);
plot(RR2,'--or','LineWidth',2,'MarkerSize',8);
xlabel('iterations','Fontsize',10);
ylabel('RQ values \rho(z_k)','Fontsize',10);
h_legend=legend('simple iter', 'Alg.1', 'Alg.2');
set(h_legend,'FontSize',10);
xlim([0,15]);
title('Situation 2');

figure
semilogy(abs(RR2t-RR2t(end)),'--xk','LineWidth',2,'MarkerSize',8);
hold on;
semilogy(abs(RR2-RR2(end)),'--or','LineWidth',2,'MarkerSize',8);
xlabel('iterations','Fontsize',10);
ylabel('|\rho(z_k) -\rho(z_*)|','Fontsize',10);
h_legend=legend('Alg.1', 'Alg.2');
xlim([0,15]);
set(h_legend,'FontSize',10);
title('Situation 2');


if ~DO_FEATURE_EX, return; end

% ++++++++++++++++++++++++++++++
% --- Feature extraction testing
% ++++++++++++++++++++++++++++++

% ** RCSP
zz1 = z1; zz2=z2;
% feature extraction
Tr1 = feature_ext(zz1, zz2, Train1);
Tr2 = feature_ext(zz1, zz2, Train2);
Ts1 = feature_ext(zz1, zz2, Tst1);
Ts2 = feature_ext(zz1, zz2, Tst2);
% LDA classifier
w = lda_train(Tr1, Tr2);
% compute centroid
c1 = mean(Tr1')*w;
c2 = mean(Tr2')*w;
d1 = var(Tr1'*w);
d2 = var(Tr2'*w);

Tsw1 = w'*Ts1;
NumMiss = nnz(abs(Tsw1-c1)/d1 > abs(Tsw1-c2)/d2);
Tsw2 = w'*Ts2;
NumMiss = NumMiss + nnz(abs(Tsw2-c2)/d2 > abs(Tsw2-c1)/d1);
MissRate = NumMiss/100;
doplot(Tr1, Tr2, Ts1, Ts2, w, 'NEP2');
disp(['NEP2:    miss = ', num2str(NumMiss),'  missrate = ',num2str(MissRate)]);
Ts10 = Ts1; Ts20 = Ts2;

% ** Simple iter
zz1 = z1s; zz2=z2s;
% feature extraction
Tr1 = feature_ext(zz1, zz2, Train1);
Tr2 = feature_ext(zz1, zz2, Train2);
Ts1 = feature_ext(zz1, zz2, Tst1);
Ts2 = feature_ext(zz1, zz2, Tst2);
% LDA classifier
w = lda_train(Tr1, Tr2);
% compute centroid
c1 = mean(Tr1')*w;
c2 = mean(Tr2')*w;
d1 = var(Tr1'*w);
d2 = var(Tr2'*w);

Tsw1 = w'*Ts1;
NumMiss = nnz(abs(Tsw1-c1)/d1 > abs(Tsw1-c2)/d2);
Tsw2 = w'*Ts2;
NumMiss = NumMiss + nnz(abs(Tsw2-c2)/d2 > abs(Tsw2-c1)/d1);
MissRate = NumMiss/100;
doplot(Tr1, Tr2, Ts1, Ts2, w, 'Simple iter');
disp(['Simple iter    miss = ', num2str(NumMiss),'  missrate = ',num2str(MissRate)]);

figure
plot(Ts1(1,:), Ts1(2,:), '+b');
hold on;
plot(Ts2(1,:), Ts2(2,:), 'ob');
plot(Ts10(1,:), Ts10(2,:), '+r');
plot(Ts20(1,:), Ts20(2,:), 'or');

return;

%
% ****************************** AUX FUN ********************************
%

% ---------------------------------------------------- %
%   doplot: plotting function 
% ---------------------------------------------------- %
function [] = doplot(Tr1, Tr2, Ts1, Ts2, w, method)
    figure;
    subplot(2,2,1)
    plot(Ts1(1,:), Ts1(2,:), '+r');
    hold on;
    plot(Ts2(1,:), Ts2(2,:), 'ob');
    title(['Test feature ', method]);
    legend('Class A', 'Class B');

    subplot(2,2,2)
    PP = w*w'*Ts1;
    plot(PP(1,:), PP(2,:), '+r');
    hold on;
    PP = w*w'*Ts2;
    plot(PP(1,:), PP(2,:), 'ob');
    title('LDA plane');
    legend('Class A', 'Class B');

    subplot(2,2,3)
    plot(Tr1(1,:), Tr1(2,:), '+r');
    hold on;
    plot(Tr2(1,:), Tr2(2,:), 'ob');
    title(['Training feature ', method]);
    legend('Class A', 'Class B');

    subplot(2,2,4)
    PP = w*w'*Tr1;
    plot(PP(1,:), PP(2,:), '+r');
    hold on;
    PP = w*w'*Tr2;
    plot(PP(1,:), PP(2,:), 'ob');
    title('LDA plane');
    legend('Class A', 'Class B');
end


% ---------------------------------------------------- %
%   GMAT2: compute G and tilde G
% ---------------------------------------------------- %
function [G TG info] = GMAT2(w, lam2, V2, delta, Sbar1)
    [G TG] = optmat_build(w, lam2, V2, delta);
    G = Sbar1 + G;
end

% ---------------------------------------------------- %
%   HMAT2: compute H and tilde H
% ---------------------------------------------------- %
function [H TH info] = HMAT2(w, lam1, V1, delta, Sbar1)
    [H TH] = optmat_build(w, lam1, V1, delta);
    %H = Sbar1 - H;
    [H info] = NegTrunc(Sbar1 - H); % suggested in [Kawanabe, et al., 2014]
    if ~info, disp('nonsmooth point in H matrix'), end;
    TH = -TH;
end

% ---------------------------------------------------- %
%   NegTrunc: truncate the negative eig mode in H
% ---------------------------------------------------- %
function [H issmth] = NegTrunc(H)
    [V, D] = eig(H);
    H = V*max(D,zeros(size(D)))*V';
    issmth = (min(diag(D)) >= 0);
end

% ---------------------------------------------------- %
%   csptest: feature extraction test 
% ---------------------------------------------------- %
function MissRate = csptest(zz1, zz2, Train1, Train2, Tst1, Tst2, FLAG)
    Tr1 = feature_ext(zz1, zz2, Train1);
    Tr2 = feature_ext(zz1, zz2, Train2);
    Ts1 = feature_ext(zz1, zz2, Tst1);
    Ts2 = feature_ext(zz1, zz2, Tst2);
    % LDA classifier
    w = lda_train(Tr1, Tr2);
    c1 = mean(Tr1'*w);
    c2 = mean(Tr2'*w);
    d1 = 1; d2 = 1;

    Tsw1 = w'*Ts1;
    NumMiss = nnz(abs(Tsw1-c1)/d1 > abs(Tsw1-c2)/d2);
    Tsw2 = w'*Ts2;
    NumMiss = NumMiss + nnz(abs(Tsw2-c2)/d2 > abs(Tsw2-c1)/d1);
    MissRate = NumMiss/(length(Tsw1)+length(Tsw2));
    disp([FLAG,':   miss = ', num2str(NumMiss),'  missrate = ',num2str(MissRate)]);
end
