% Example for robust LDA

clear all; close all

addpath ./rlda/
rng(0); 

% testing cases
TST = 's'; % sonar
%TST = 'i'; % ionosphere


DO_CVX = false; % run robust LDA by CVX if true
if DO_CVX && ~exist('cvx_version')
    % START UP CVX: CVX has to be installed
    disp('    example5: CVX not installed!')
    return
end
DO_NEPV1 = false; % also run NEPv1 if true

if strfind(TST, 's')
    load sonar.mat;
    sonar_label = (sonar_label == 77); % turn to 0-1 label
    DATA = [sonar_data, sonar_label]; % merge data and label for simplicity 
    ALPHA = [0.3, 0.4, 0.5, 0.6, 0.7, 0.8];
else
    load ionosphere.mat;
    ionosphere_label = (ionosphere_label == 103); % turn to 0-1 label
    DATA = [ionosphere_data, ionosphere_label]; % merge data and label for simplicity 
    % TRUNCATE FIRST TWO FEATURES.. ALMOST THE SAME.. leads to singular cov
    %DATA(:,2) = [];
    %DATA(:,1) = [];
    ALPHA = [0.10, 0.15, 0.2, 0.4, 0.6, 0.7];
end

n = size(DATA, 2) - 1;
N_SAMPLE = size(DATA, 1);
MM = []; STD = []; ITS = [];
NST = 0; 
TIME_CVX = []; TIME_NEP1 = []; TIME_NEP2 = [];

for alpha = ALPHA
    N_TRAIN = floor(N_SAMPLE*alpha);
    N_TST = N_SAMPLE - N_TRAIN;
    N_RESAMPLETIME = 100;

    R1 = []; R2 = []; R3 = [];
    for NN_TST = 1:100
        % random partition data
        RDPM = randperm(N_SAMPLE);
        DATA_TRAIN = DATA(RDPM(1:N_TRAIN),:);
        DATA_TST = DATA(RDPM(N_TRAIN+1:end),:);

        UX = []; UY = [];
        SX = []; SY = [];
        for i = 1 : N_RESAMPLETIME
            % uniformly distributed resample
            RESAMPLE_IDX= randi(N_TRAIN, N_TRAIN, 1);
            TRAIN_TMP = DATA_TRAIN(RESAMPLE_IDX,:);
            TRLABEL = TRAIN_TMP(:, end);
            TRAIN_TMP_c0 = TRAIN_TMP(TRLABEL==0,1:n);
            TRAIN_TMP_c1 = TRAIN_TMP(TRLABEL==1,1:n);
            while size(TRAIN_TMP_c0,1) <= 1 || size(TRAIN_TMP_c1,1) <= 1
                % redo if too few sample
                RESAMPLE_IDX= randi(N_TRAIN, N_TRAIN, 1);
                TRAIN_TMP = DATA_TRAIN(RESAMPLE_IDX,:);
                TRLABEL = TRAIN_TMP(:, end);
                TRAIN_TMP_c0 = TRAIN_TMP(TRLABEL==0,1:n);
                TRAIN_TMP_c1 = TRAIN_TMP(TRLABEL==1,1:n);
            end
            ux0 = mean(TRAIN_TMP_c0); % mean of class 0
            sx0 = cov(TRAIN_TMP_c0); % cov of class 0
            uy0 = mean(TRAIN_TMP_c1); % mean  of class 1
            sy0 = cov(TRAIN_TMP_c1); % cov  of class 1

            UX = [UX; ux0];
            UY = [UY; uy0];
            SX = [SX; sx0(:)']; % vectorize cov
            SY = [SY; sy0(:)']; % vectorize cov

        end

        ux = mean(UX)'; 
        uy = mean(UY)'; 
        [Px, IPx] = trinv(cov(UX)); %+ 1.0E-8*eye(n);
        Px = Px/n; IPx = IPx*n;
        [Py, IPy] = trinv(cov(UY)); %+ 1.0E-8*eye(n);
        Py = Py/n; IPy = IPy*n;

        msx = mean(SX);
        dfsx = SX - ones(N_RESAMPLETIME,1)*msx;
        dx = sqrt(max(sum(dfsx.*dfsx, 2)));
        %Sx = truncate(reshape(msx, n,n));
        Sx = reshape(msx, n,n); % + 1.0E-8*eye(n);
        %Sx = reshape(msx, n,n);

        msy = mean(SY);
        dfsy = SY - ones(N_RESAMPLETIME,1)*msy;
        dy = sqrt(max(sum(dfsy.*dfsy, 2)));
        %Sy = truncate(reshape(msy, n,n));
        Sy = reshape(msy, n,n); %+ 1.0E-8*eye(n);
        %Sy = reshape(msy, n,n);

        % ++++++++++++++++++++++++++++++
        % --- solve by cvx
        % ++++++++++++++++++++++++++++++
        if DO_CVX
            tic
            [w, optr] = rlda_cvx(Sx, Sy, dx, dy, ux, uy, Px, Py);
            t0 = toc;
            TIME_CVX = [TIME_CVX, t0];
        else 
            w = pinv(Sx+Sy)*(ux-uy);
        end

        % ++++++++++++++++++++++++++++++
        % --- solve by NEP
        % ++++++++++++++++++++++++++++++
        P0 = Sx+Sy+(dx+dy)*eye(n);

        Evalfw = @(z) FW(z, ux, uy, IPx, IPy);
        EvalTH = @(z) HMAT(z, ux, uy, IPx, IPy);
        EvalTG = @(z) GMAT(z, P0);
        z0 = pinv(Sx+Sy)*(ux-uy);

        % check feasibility..
        t00 = 0;
        if norm(EvalTH(z0)) == 0
            tic;
            z0 = rlda_nepv_start(ux, uy, Px, Py, IPx, IPy);
            t00 = toc;
            disp('starting vec search')
            NST = NST + 1; 
        end


        if norm(EvalTH(z0)) == 0
            return;
        end
            
        warning ('off','all');
        % Run SCF for NEP 1st 
        if DO_NEPV1 
            tic
            [z2, r2, ~, its2, resd2] = scfsls_rlda(P0, Evalfw, z0, true, 1.0E-3);
            t2 = toc + t00;
            TIME_NEP1 = [TIME_NEP1, t2];
        end
        % Run SCF for NEP 2nd 
        tic
        [z3, r3, ~, its3, resd3] = scfls(EvalTG, EvalTH, z0, true, 1.0E-8);
        t3 = toc + t00;
        ITS = [ITS, its3];
        TIME_NEP2 = [TIME_NEP2, t3];

        % ++++++++++++++++++++++++++++++
        % --- results..
        % ++++++++++++++++++++++++++++++
        fprintf( '\n ************************************************\n ')
        rho = @(z) z'*EvalTG(z)*z/(z'*EvalTH(z)*z);
        if DO_CVX 
            disp( ' -- CVX ' )
            fprintf(' cvx opt: rq(z) = %6.5e,\t fval = %6.5e,\t \t time = %3.2fs\n', rho(w),  optr, t0);
        end

        disp( ' -- NEP ' )
        if DO_NEPV1 
            fprintf(' 1st NEP: rq(z) = %6.5e,\t its = %3i, \t resd = %2.1e, \t time = %3.2fs \n', rho(z2), its2, resd2(end), t2);
        end
        fprintf(' 2nd NEP: rq(z) = %6.5e,\t its = %3i, \t resd = %2.1e, \t time = %3.2fs \n', rho(z3), its3, resd3(end), t3);

        % ++++++++++++++++++++++++++++++
        % --- classification
        % ++++++++++++++++++++++++++++++

        TRLABEL = DATA_TRAIN(:, end);
        Tr1 = DATA_TRAIN(TRLABEL==0,1:n);
        Tr2 = DATA_TRAIN(TRLABEL==1,1:n);
        TSLABEL = DATA_TST(:, end);
        Ts1 = DATA_TST(TSLABEL==0,1:n);
        Ts2 = DATA_TST(TSLABEL==1,1:n);

        MissRate = classification( w, Tr1', Tr2', Ts1', Ts2');
        R1 = [R1,1-MissRate];

        if DO_NEPV1 
            MissRate = classification( z2, Tr1', Tr2', Ts1', Ts2');
            R2 = [R2,1-MissRate];
        end

        MissRate = classification( z3, Tr1', Tr2', Ts1', Ts2');
        R3 = [R3,1-MissRate];

    end

    MM = [MM; mean([R1',R3',R2'])];
    STD = [STD; std([R1',R3',R2'])];
end 

errorbar(ALPHA*100, MM(:, 1), STD(:,1), STD(:,1), '--b');
hold on;
errorbar(ALPHA*100, MM(:, 2), STD(:,2), STD(:,2), 'r');
xlabel('\alpha')
ylabel('TSA')
if DO_CVX 
    legend('CVX', 'NEPv', 'Location', 'SouthEast');
else
    legend('LDA', 'RLDA', 'Location', 'SouthEast');
end
xlim([0,100])
ylim([0.5,0.9])

disp('averaged its:')
mean(ITS)

disp('averaged time:')
mean([TIME_CVX', TIME_NEP1', TIME_NEP2'])

disp('number of starting value reset:')
NST

return



%
% ****************************** AUX FUN ********************************
%
% ---------------------------------------------------- %
%   GMAT: compute G and tilde G
% ---------------------------------------------------- %
function [G, TG] = GMAT(w, M)
    G = M;
    TG = 0*M;
end


% ---------------------------------------------------- %
%   HMAT: compute H and tilde H
% ---------------------------------------------------- %
function [H, TH] = HMAT(w, ux, uy, IPx, IPy)
    n = length(ux);
    dxy = ux-uy;
    ipxw = IPx*w; wpxw = sqrt(w'*ipxw);
    ipyw = IPy*w; wpyw = sqrt(w'*ipyw);
    if abs(w'*dxy) <= (wpxw + wpyw )
        H = zeros(length(dxy));
        TH = zeros(length(dxy));
        return
    end
    fw = dxy - sign(w'*dxy)*(ipxw/wpxw + ipyw/wpyw);
    H = fw*fw'; H = (H+H')/2;
    TH = -sign(w'*dxy)*( (IPx - ipxw*ipxw'/ (wpxw)^2)/wpxw + (IPy - ipyw*ipyw'/ (wpyw)^2)/wpyw);
    TH = (TH + TH')/2;
    TH = (w'*fw) * TH;
end


% ---------------------------------------------------- %
%   FW: compute fw(z), s.t., H = f*f'
% ---------------------------------------------------- %
function fw = FW(w, ux, uy, IPx, IPy)
    n = length(ux);
    dxy = ux-uy;
    ipxw = IPx*w; wpxw = sqrt(w'*ipxw);
    ipyw = IPy*w; wpyw = sqrt(w'*ipyw);
    if abs(w'*dxy) <= (wpxw + wpyw )
        fw = zeros(length(dxy),1);
        return
    end
    fw = dxy - sign(w'*dxy)*(ipxw/wpxw + ipyw/wpyw);
end
 

% ---------------------------------------------------- %
%   CLASSIFICATION: compute LDA classification rate 
% ---------------------------------------------------- %
function MissRate = classification( w, Tr1, Tr2, Ts1, Ts2)
    c1 = mean(Tr1'*w);
    c2 = mean(Tr2'*w);
    d1 = 1; d2 = 1;
    Tsw1 = w'*Ts1;
    NumMiss = nnz(abs(Tsw1-c1)/d1 > abs(Tsw1-c2)/d2);
    Tsw2 = w'*Ts2;
    NumMiss = NumMiss + nnz(abs(Tsw2-c2)/d2 > abs(Tsw2-c1)/d1);
    MissRate = NumMiss/(length(Tsw1)+length(Tsw2));
    %disp([' miss = ', num2str(NumMiss),'  missrate = ',num2str(MissRate)]);
end


% ---------------------------------------------------- %
%   TRUNCATE: truncate small singular values in M
% ---------------------------------------------------- %
function M = truncate(M)
    tol = 1.0E-8;
    [V,E] = eig(M); %M = V*E*V'
    E(E<max(E)*tol) = 0;
    M = V*E*V'; M = (M+M')/2;
end


% ---------------------------------------------------- %
%   TRINV: reset small singular values of M, return pseudo inverse 
% ---------------------------------------------------- %
function [M, IM] = trinv(M)
    tol = 1.0E-8;
    [V,E] = eig(M); %M = V*E*V'
    e = diag(E);
    
    % set small e to max(e)*tol
    e(e<max(e)*tol) = max(e)*tol;
    IM = V*diag(e)*V'; IM = (IM+IM')/2;
    E = pinv(diag(e));
    M = V*E*V'; M = (M+M')/2;
    %    IM = M;
    %    M = pinv(M);
    %    M = (M+M')/2;
end
