function [V1, lam1, Sbar1, V2, lam2, Sbar2, Tst1, Tst2, Train1, Train2] = datagen()
% function datagen generates random experiments data for the linear mixing model.

NUMTRIALS_TRAIN = 50;
TRAIN_SAMPLE_PERTRIAL = 200;
TST_SAMPLE_PERTRIAL = 200;

EPSLN_TR = 2;
EPSLN_TST = 30;

% RAND ROTATION
[A,~] = qr(rand(10));
% SAMPLE DATA
for i = 1:NUMTRIALS_TRAIN
    % sample covariance in trial i
    X1 = []; X2 = [];
    Y1 = []; Y2 = [];
    for j = 1:TRAIN_SAMPLE_PERTRIAL
        s1 = [randn*sqrt(1.8); randn*sqrt(0.6); randn(8,1)];
        dice = rand;
        XX = A*s1 + randn*sqrt(EPSLN_TR);
        X1 = [X1, XX];

        s1 = [randn*sqrt(1.8); randn*sqrt(0.6); randn(8,1)];
        dice = rand;
        XX = A*s1 + randn*sqrt(EPSLN_TST);
        Y1 = [Y1, XX];

        s2 = [randn*sqrt(0.2); randn*sqrt(1.4); randn(8,1)];
        dice = rand;
        XX = A*s2 + randn*sqrt(EPSLN_TR);
        X2 = [X2, XX];

        s2 = [randn*sqrt(0.2); randn*sqrt(1.4); randn(8,1)];
        dice = rand;
        XX = A*s2 + randn*sqrt(EPSLN_TST);
        Y2 = [Y2, XX];
    end
    S1{i} = cov(X1');%X1 * X1' / (SAMPLE_PERTRIAL - 1);
    S2{i} = cov(X2');%X2 * X2' / (SAMPLE_PERTRIAL - 1);

    Train1{i} = X1; Train2{i} = X2;
    Tst1{i} = Y1; Tst2{i} = Y2;
end

[V1, Sbar1, lam1] = PCA_V(S1, 10);
[V2, Sbar2, lam2] = PCA_V(S2, 10);
return

% ---------------------------------------------------- %
%   AUXFUN: PCA for the V matrices
% ---------------------------------------------------- %
function [V, Sbar1, lam] = PCA_V(S1, n)
%  truncate the tailing n-k1 singular values.
%   you can also set tol to truncate singular values less than |S1|*tol.

%compute Vs by SVD
S1_mat = cell2mat(S1);
S1_vec = reshape(S1_mat, n*n, []);
Sbar1 = mean(S1_vec')'; 
Sbar1 = reshape(Sbar1, n,n);

CovM = cov(S1_vec');
[VV, E] = eig(CovM);
lam1 = diag(E);

%tol = eps;
%k1 = nnz(lam1 > tol*lam1(end));
k1 = 10; % fixed 

lam = mat2cell(lam1(end-k1+1:end)', 1, ones(1,k1));
VV = VV(:, end-k1+1:end);
V1 = reshape(VV, n, []);

% symmetrize (CAN ALSO USE TRIL PART TO REPRESENT MAT)
V1T = V1'; V1T = reshape(V1T, n, k1, n); 
V1T = permute(V1T, [1, 3, 2]); 
V1T = reshape(V1T, n, n*k1);
V1 = (V1+V1T)/2;
V = mat2cell(V1, n, n*ones(1,k1));
return






