function [P dP] = optmat_build(w, lam, V, delta)
% function optmat_build compute the optimal perturbation matrix, for
%   generating the matrices 
%       G(w) = Sigma + optmat_build(w, lam, V, delta)
%       H(w) = Sigma - optmat_build(w, lam, V, delta)
%
%   Input: lam{i} = lambda_i, V{i} = V_i.
%
P = zeros(size(V{1}));

% evaluate wtVw{i} = w'*V{i}*w
wtVw_fun = @(Vi) w'*Vi*w;
wtVw = cellfun(wtVw_fun, V);
nn = sqrt( wtVw.^2*cell2mat(lam)' );

for i = 1:length(V);
    alpha = delta*(lam{i}*wtVw(i))/nn;
    P = P+alpha*V{i};
end

if nargout < 2
    return;
end

% evaluate dP
S1 = 0; S2 = 0; S3 = 0;
for i = 1:length(V)
    S1 = S1 + (V{i}'*w)*(w'*V{i})*lam{i};
    S2 = S2 + wtVw(i)*V{i}'*w*lam{i};
end

dP = 2*delta*(S1/nn-S2*S2'/nn^3);


return
