function w = rlda_nepv_start(ux, uy, Px, Py, IPx, IPy)
% Find initial vector for NEPV, that is, to find w, s.t.,
%   |w'(ux-uy)| > sqrt(w'*Px*w) + sqrt(w'*Py*w)
%   The problem is not feasible if w has no solution.
% 
if ~exist('cvx_version')
    error('RLDA_NEPV_START: CVX not installed!');
    return
end

n = length(ux);
d = ux - uy;
%L = chol(Py, 'lower');
L = chol(IPy, 'lower');
Pxy = (L'*Px)*L;
hd = L\d;

% Solve QP
%   min     u'u 
%   s.t.    (u-hd)'*Pxy*(u-hd) <= 1

q1 = -2*Pxy*hd;
r1 = hd'*Pxy*hd - 1;
cvx_begin quiet
    variable u(n)
    dual variables lam1 
    minimize( norm(u))
    lam1: quad_form(u,Pxy) + q1'*u + r1 <= 0;
cvx_end

obj1 = sqrt(cvx_optval);
if obj1 < 1
    disp(' not feasible ');
end

tw = u/norm(u);
w = L'\tw;

return




