function [z1, r, R, i, Resd] = scfitr(EvalG, EvalH, z1, tol)
% Fixed-point simple iterative scheme, equivalent to SCF iteration 
%   for solving the minimum eigenvalue of G(z)*z = lambda*H(z)*z.

% INPUT:
%   EvalG   function handle G(z) = EvalG(z)
%   EvalH   function handle H(z) = EvalH(z)
%   z1      initial vector
%   tol     tolerance


% OUTPTU:
%   z1 :    optimizer
%   r :     rho values of each iteration
%   R :     each column contains the z from each iteration.
%   i :     iteration number
%   Resd:   always 0 (dummy output).

maxit = 30;
if nargin<4, tol   = 1.0E-6; end
R = [];
r = [];
for i = 1: maxit
   G = EvalG(z1);
   H = EvalH(z1);
   rr = z1'*G*z1 / (z1'*H*z1);   
   r = [r, z1'*G*z1/(z1'*H*z1)];
   R = [R,z1];
   % Convergence test 
   if i > 1 && norm(r(end-1)-rr)< tol*rr
        break
   end
   % SCFI updates 
   z0 = z1;
   [V,E] = eig(G,H);
   [C,IDX] = min(diag(E));
   z1 = V(:,IDX);
   z1 = z1/norm(z1);
end
[~,idx] = min(r);
z1 = R(:,idx);
Resd = 0; 
return;
